/*                                                                */
/*                 S A 9 6 0 7 P   D e m o  s o f t w a r e       */
/*                                                                */
/*                          SAMES Pty Ltd                         */
/*                        REV1.00   9-June-2000                   */
/*                                                                */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <conio.h>

#define LPT1  0x378 //= output, LPT1+1 = input
#define LPT2  0x278 //= output, LPT2+1 = input

#define setbit(x) (portval|x)
#define clearbit(x) (portval&(~x))

#define OPTO 1

// look at Sendbit for inverted bits due to the optocouplers
#define TEST   0x01      //pin 2
#define DIRO   0x02      //pin 3
#define VDD2   0x20      //pin 7

#define SDO    0x04      //pin 4
#define VDD    0x40      //pin 8
#define SCL    0x80      //pin 9
#define SDI    0x20      //pin 12
#define SADR     12      //Start Address for data in EE
#define PPDM 642000      //PPDM from Analog

const int write=0x00;    // To set the l.s bit of the byte
const int read = 0x01;

long delay_time;
int  prt;

int           port;
unsigned char p1;        // p1 is 8 bits 0 to 255
unsigned char portval;   // Glabal variable to keep the LPT status

char  VAL[32],Cled,Cpw,Cres,Csel;

// Default values for program start up
int   VOLT=230, CURR=40, PW=140, CRES=10, CONST=6400,
      Kc1=511, Kc2=511, Ne=2, Kr=51, CSEL=0;
float ERR1=0,ERR2=0, ELEAK;

void Calculate()
{
  float Ktot, Krx, Corr;

  Ktot = (float)PPDM/VOLT/CURR/6400*3600*1000;
  Krx  = Ktot/511;
  Kr   = Krx;
  if (Kr>256) Kr=256;
  if (Kr<2) Kr=2;

  Kc1  = Krx/Kr*512*(1+ERR1/100);
  Kc2  = Krx/Kr*512*(1+ERR2/100);

  if (CONST==6400) Cled=0;
  if (CONST==3200) Cled=1;
  if (CONST==0) Cled=2;

  if (PW>100) Cpw=0x40;
  else
    Cpw=0;
  if (CRES==1) Cres=0;
  if (CRES==10) Cres=1;
  if (CRES==100) Cres=2;

  if (CSEL==0) Csel=0;
  if (CSEL==1) Csel=12;
  if (CSEL==2) Csel=4;

  ELEAK=1.8*(Kc1+Kc2)/2/PPDM*CURR/(1+Ne);
}

void Box(char x, char y)
{
  int x0, y0, i;

  x0 = wherex();
  y0 = wherey();

  for (i=1; i<x; i++)
  {
    gotoxy(x0+i,y0);   printf("%c",(char)196);
    gotoxy(x0+i,y0+y); printf("%c",(char)196);
  }
  for (i=1; i<y; i++)
  {
    gotoxy(x0,y0+i);   printf("%c",(char)179);
    gotoxy(x0+x,y0+i); printf("%c",(char)179);
  }

  gotoxy(x0,y0);     printf("%c",(char)218);
  gotoxy(x0+x,y0);   printf("%c",(char)191);
  gotoxy(x0,y0+y);   printf("%c",(char)192);
  gotoxy(x0+x,y0+y); printf("%c",(char)217);
}
// Draw channel 1 box and text
void Channel1(void)
{
  int x=2, y=2;

  gotoxy(x+6,1);
  printf("Live Power");

  gotoxy(x,y);
  Box(20, 3);

  gotoxy(x+10,y+3);
  printf("%c",(char)194);

  gotoxy(x+3,y+1);
  printf("Chnl1 Balance");
  gotoxy(x+3,y+2);
  printf("%c%d",(char)246,Kc1);
}

// Draw channel 2 box and text
void Channel2(void)
{
  int x=31, y=2;

  gotoxy(x+4,1);
  printf("Neutral Power");

  gotoxy(x,y);
  Box(20, 3);

  gotoxy(x+10,y+3);
  printf("%c",(char)194);

  gotoxy(x+3,y+1);
  printf("Chnl2 Balance");
  gotoxy(x+3,y+2);
  printf("%c%d",(char)246,Kc2);
}

// Draw earthleak box and text
void EarthLeak(void)
{
  int x=59, y=2;

  gotoxy(x,y);
  Box(20, 3);

  gotoxy(x+9,y+3);
  printf("%c",(char)194);

  gotoxy(x+2,y+1);
  printf("Earth Leak Comp");
  gotoxy(x+3,y+2);
  printf("%1.4fA",ELEAK);
}

// Draw channel select box and text
void ChnlSel(void)
{
  int i, x=2, y=6;

  gotoxy(x,y);
  Box(77, 2);

  gotoxy(x+24,y+2);
  printf("%c",(char)194);
  gotoxy(x+24,y+3);
  printf("%c",(char)195);

  for (i=1; i<18; i++)
  {
    gotoxy(x+24+i,y+3);
    printf("%c",(char)196);
  }
  printf("%c",(char)191);

  for (i=1; i<6; i++)
  {
    gotoxy(x+24+18,y+3+i);
    printf("%c",(char)179);
  }
  gotoxy(x+10,y+8);
  printf("%c",(char)218);
  for (i=1; i<26; i++)
  {
    gotoxy(x+10+i,y+8);
    printf("%c",(char)196);
  }
  printf("%c",(char)191);
  gotoxy(x+24,y+8);
  printf("%c",(char)193);

  gotoxy(x+15,y+1);
  printf("Channel Select ");
  if (CSEL==0)
    printf("auto");
  else
    printf("%d",CSEL);
}

// Draw rated conditions box and text
void Rated(void)
{
  int x=16, y=10;

  gotoxy(x,y);
  Box(20, 3);

  gotoxy(x+10,y+3);
  printf("%c",(char)194);

  gotoxy(x+2,y+1);
  printf("Rated Condition");
  gotoxy(x+2,y+2);
  printf("%c%d (%dV/%dA)",(char)246,Kr,VOLT,CURR);
}

// Draw counter resulution box and text
void CntResol(void)
{
  int x=2, y=15;

  gotoxy(x,y);
  Box(20, 3);

  gotoxy(x+10,y+3);
  printf("%c",(char)194);

  gotoxy(x+2,y+1);
  printf("Counter Resolut.");
  gotoxy(x+3,y+2);
  printf("%d",CRES);
}

// Draw LED constant box and text
void LEDConst(void)
{
  int i, x=31, y=15;

  gotoxy(x,y);
  Box(20, 3);

  gotoxy(x+10,y+3);
  printf("%c",(char)194);
  for (i=1; i<5; i++)
  {
    gotoxy(x+10,y+3+i);
    printf("%c",(char)179);
  }
  gotoxy(x+9,y+8);
  printf("LED");

  gotoxy(x+1,y+1);
  printf(" LED Constant ");

  gotoxy(x+3,y+2);
  printf("%d",CONST);
}

// Draw counter pulse width box and text
void CntPw(void)
{
  int x=2, y=19;

  gotoxy(x,y);
  Box(20, 3);

  gotoxy(x+8,y+3);
  printf("%c",(char)194);
  gotoxy(x+12,y+3);
  printf("%c",(char)194);

  gotoxy(x+7,y+4);
  printf("MOP");
  gotoxy(x+11,y+4);
  printf("MON");

  gotoxy(x+2,y+1);
  printf("Counter Pulswidth");
  gotoxy(x+3,y+2);
  printf("%dms",PW);
}

// Set up the menu
void Menu(char Cpos)
{
  int x=55, y=10;

  gotoxy(x,y);
  textcolor(BLACK); textbackground(LIGHTGRAY);
  cprintf("SA9607P Meter Parameters");
  textcolor(LIGHTGRAY); textbackground(BLACK);
  gotoxy(x,y+2);
  printf("Vnom          = %dV    ",VOLT);
  gotoxy(x,y+3);
  printf("Imax          = %dA    ",CURR);
  gotoxy(x,y+4);
  printf("LED Const.    = %dp/kWh",CONST);
  gotoxy(x,y+5);
  printf("Counter Resol.= %dp/kWh",CRES);
  gotoxy(x,y+6);
  printf("Counter PW    = %dms   ",PW);
  gotoxy(x,y+7);
  printf("Channel1 Err  = %2.2f% ",ERR1);
  gotoxy(x,y+8);
  printf("Channel2 Err  = %2.2f% ",ERR2);
  gotoxy(x,y+9);
  printf("Channel Sel.  = %d     ",CSEL);
  gotoxy(x,y+10);
  printf("Earth Leak    = %d     ",Ne);

  gotoxy(x-8,y+12);
  printf("'r'ead, 'w'rite, 'l'istmem");
  gotoxy(x-8,y+13);
  printf("'t'rig-load 'q'uit");

  gotoxy(72-1,y+1+Cpos);
}

// Keep cursor in bounds
char CursorPos(int c)
{
  int x=72, y=11;
  static char Cpos;

  c = c>>8;
  if (c == 80)
  {
    Cpos++;
    if (Cpos>9) Cpos=9;
  }
  if (c == 72)
  {
    Cpos--;
    if (Cpos<1) Cpos=1;
  }
  gotoxy(x-1,y+Cpos);
  return(Cpos);
}

// Draw the complete block diagram
void BlockDia()
{
  clrscr();
  Channel1();
  Channel2();
  EarthLeak();
  ChnlSel();
  Rated();
  CntResol();
  CntPw();
  LEDConst();
}

// Change the parameters accoding to input
void GetPar(char Cpos)
{
  char y=11;

  if (Cpos==1)
  {
    printf("Volt      ");
    gotoxy(71,y+Cpos);
    scanf("%d",&VOLT);
  }
  if (Cpos==2)
  {
    printf("Amp       ");
    gotoxy(71,y+Cpos);
    scanf("%d",&CURR);
  }
  if (Cpos==3)
  {
    printf("p/kWh     ");
    gotoxy(71,y+Cpos);
    scanf("%d",&CONST);
    if(CONST<1) CONST=0;
    else
      if (CONST<4000) CONST=3200;
      else CONST=6400;
  }
  if (Cpos==4)
  {
    printf("p/kWh     ");
    gotoxy(71,y+Cpos);
    scanf("%d",&CRES);

    if(CRES>100) CRES=100;
    if(CRES<1)   CRES=1;
    if((CRES>1)&&(CRES<100)) CRES=10;
  }
  if (Cpos==5)
  {
    printf("ms        ");
    gotoxy(71,y+Cpos);
    scanf("%d",&PW);
    if(PW>100) PW=142; else PW=71;
  }
  if (Cpos==6)
  {
    printf("Err/%     ");
    gotoxy(71,y+Cpos);
    scanf("%f",&ERR1);
  }
  if (Cpos==7)
  {
    printf("Err/%     ");
    gotoxy(71,y+Cpos);
    scanf("%f",&ERR2);
  }
  if (Cpos==8)
  {
    printf("1,2,0=auto");
    gotoxy(71,y+Cpos);
    scanf("%d",&CSEL);
    if (CSEL>1) CSEL=2;
    if (CSEL<1) CSEL=0;
  }
  if (Cpos==9)
  {
    printf("0,1,2");
    gotoxy(71,y+Cpos);
    scanf("%d",&Ne);
    if (Ne>2) Ne=2;
    if (Ne<0) Ne=0;
  }
}

// Print register valuse as well as binary values
void PrintVal(char val)
{
  unsigned int i, temp;
  // temp used to print charc correcly
  temp=(unsigned int)val&0x00FF;
  printf("%8u\t0x%x\t",temp,temp);
  for (i=0; i<8; i++)
  {
    if (val<<i&0x80)
      printf("1");
    else
      printf("0");
   if(i==3)
   printf(".");
  }
}

//Do memory listing
void MemList(void)
{
  int i;

  clrscr();
  printf("ADDR     DEC    HEX  BIN[7:0]\n");
  for (i=0; i<32; i++)
  {
    printf("%4d",i); PrintVal(VAL[i]);
    if(i==12)
      printf(" Channel 1 balance LSB");
    if(i==14)
      printf(" Channel 1 balance MSB");
    if(i==16)
      printf(" Channel 2 balance LSB");
    if(i==18)
      printf(" Channel 2 balance MSB");
    if(i==20)
      printf(" Rated Condition");
    if(i==22)
      printf(" LED Pulse-rate");
    if(i==24)
      printf(" Counter Pulse-width and Resolution");
    if(i==26)
      printf(" Earth Leakage Compensation and channel select");
    printf("\n");

    if (i==20)
    {
      printf("Press any key to continue");
      while (!bioskey(0));
      printf("\nADDR     DEC    HEX  BIN[7:0]\n");
    }
  }
  printf("press any key to continue");
  while (!bioskey(0));
  printf("\n");
}
// Generate memory map and checksum bits
void MemCode(void)
{
  VAL[0]=0;
  VAL[1]=~VAL[0];
  VAL[2]=0;
  VAL[3]=~VAL[2];
  VAL[4]=0;
  VAL[5]=~VAL[4];
  VAL[6]=0;
  VAL[7]=~VAL[6];
  VAL[8]=0;
  VAL[9]=~VAL[8];
  VAL[10]=0;
  VAL[11]=~VAL[10];

  VAL[12]=((Kc1-1)&0xff);        //channel1 bal lsb
  VAL[13]=~VAL[12];
  VAL[14]=(((Kc1-1)>>8)&0x03);   //channel1 bal msb
  VAL[15]=~VAL[14];

  VAL[16]=((Kc2-1)&0xff);        //channel2 bal lsb
  VAL[17]=~VAL[16];
  VAL[18]=(((Kc2-1)>>8)&0x03);   //channel2 bal msb
  VAL[19]=~VAL[18];

  VAL[20]=((Kr-1)&0xff);         //rated condition
  VAL[21]=~VAL[20];
  VAL[22]=(Cled&0x03);           //LED Const
  VAL[23]=~VAL[22];
  VAL[24]=(Cpw|Cres);            //counter puls width
  VAL[25]=~VAL[24];

  VAL[26]=(Csel|Ne);             //earth leak compensation
  VAL[27]=~VAL[26];
  VAL[28]=0;
  VAL[29]=~VAL[28];
  VAL[30]=0;
  VAL[31]=~VAL[30];
}

// Write a byte to LPT
void SendBit(unsigned char newvalue)
{
  int n;
  // Invert bits for opto couplers, not stored in portval
  newvalue^=SDO;
  outport(prt,newvalue);
  newvalue^=SDO;
  portval=newvalue;   // remember the status of the port
  for(n=0;n<delay_time;n++); // Delay for slow optocouplers
}

// IIC start condition
void start()
{
  SendBit(clearbit(SCL));   // send a clock signal
  if(OPTO)
    SendBit(setbit(SDO));   // set sdo high

  SendBit(setbit(SCL));     // send a clock signal
  SendBit(setbit(SDO));     // set sdo high
  SendBit(clearbit(SDO));   //set sdo low
}

// IIC stop condition
void stop()
{
  SendBit(clearbit(SCL));   // send a clock signal
  SendBit(setbit(SCL));     // send a clock signal
  SendBit(clearbit(SDO));   // set sdo low
  SendBit(setbit(SDO));
  if(OPTO)
    SendBit(clearbit(SCL));   // Release the SCL line
  else
    SendBit(clearbit(SDO));   // Release the SDO line
}

int send_ack()
{
  SendBit(clearbit(SCL));  //scl low
  SendBit(clearbit(SDO));
  SendBit(setbit(SCL));    //scl high
  return(1);
}

// IIC ACK condition
int receive_ack()
{
  int flag;

  SendBit(clearbit(SCL));  //set the clock
  SendBit(setbit(SDO));
  SendBit(setbit(SCL));    //set the clock

  if((inportb(prt+1))&SDI)
    flag=1;
  else
    flag=0;

  return(flag);
}

// IIC command formatting
void command(int value)
{
  int counter,n;
  int mask,and_mask;
  mask=0x80;
  value|=0xa0;   // this puts the header for the adress 1010
  for(counter=0;counter<8;counter++)
  {
    and_mask=(mask&value);

    SendBit(clearbit(SCL));     // set the clock low
    if(and_mask==mask)
      SendBit(setbit(SDO));       // set sdo high
    else
      SendBit(clearbit(SDO));     // set sdo low
    SendBit(setbit(SCL));       //set clock high
    value=value<<1;
  }
}

// IIC byte send
int Byte_Send(char value,int mask)
{
  int counter,bit;
  for(counter=0;counter<8;counter++)
  {
    SendBit(clearbit(SCL));
    if ((value<<counter)&mask)
      SendBit(setbit(SDO));   // set sdo high
    else
      SendBit(clearbit(SDO));  // set sdo low
    SendBit(setbit(SCL));
  }
  return(value);
}

// Write complete byte
void ByteWrite(char address,char value)
{
  int flag;
  delay(5);
  start();
  command(write);
  receive_ack();
  SendBit(clearbit(SDO));
  Byte_Send(address,0x80);
  receive_ack();
  SendBit(clearbit(SDO));
  Byte_Send(value,0x80);
  flag=receive_ack();
  if(flag==0)
  {
    SendBit(clearbit(SDO));
    stop();
  }
  else
    printf("Acknowledge error writing to address %d \n",address);
}

// Trigger a device relad
void TrigReload(void)
{
  SendBit(setbit(SDO)); // Dont drive SDO
  delay(1);
  SendBit(setbit(DIRO));
  SendBit(setbit(TEST));
  delay(10);
  SendBit(clearbit(TEST));
  SendBit(clearbit(DIRO));
}

void Write2EE(void)
{
  char i;
  clrscr();
  printf("   Writing values to the EEPROM\n ");
  for (i=0; i<32; i++)
    ByteWrite(i,VAL[i]);
  if(OPTO)
  {
    SendBit(clearbit(SCL)); // Dont drive SCL
    SendBit(clearbit(SDO)); // Dont drive SDO
  }
}

char Get_Byte()
{
  int counter;
  char fromport=0;
  int bit;

  for(counter=0;counter<8;counter++)
  {
    fromport<<=1;
    SendBit(clearbit(SCL));
    SendBit(setbit(SDO));
    SendBit(setbit(SCL));

    if((inportb(prt+1))&SDI)
      fromport|=0x1;     // Set bit
    else
      fromport&=~0x01;   // Clear Bit
  }
  return(fromport);
}

// IIC sequencial read
void SQ_read()
{
  int counter,address;
  unsigned int ans;
  int no_bytes;
  address =0;
  no_bytes=32;
  clrscr();

  start();
  command(write);
  receive_ack();
  SendBit(clearbit(SDO));
  Byte_Send(address,0x80);
  receive_ack();

  start();
  command(read);
  receive_ack();
  clrscr();
  for(counter=0;counter<(no_bytes-1);counter++)
  {
    VAL[counter]=Get_Byte();
    send_ack();
  }
  VAL[counter] = Get_Byte();
  SendBit(clearbit(SDO));
  stop();
}

void main (int argc,char *argv[])
{
  int port, choice;
  unsigned long result;
  int Key, E;
  char  ch=' ', Cpos;

  if (argc==3)
  {
    delay_time=atol(argv[2]);
    prt=atol(argv[1]);
    if(prt==1)
      prt=LPT1;
    else
      prt=LPT2;
  }
  else
  {
    printf("\nusage : sa9607p <portnum> <SCL delay>\n");
    printf("        <portnum>   =   1 or 2 (lpt port number)\n");
    printf("        <SCL delay> = SW loop to delay the SCL line\n");
    printf("                      use 1000 for a 486dx266      \n");
    exit(0);
  }

  portval=0;                 // Clear the paralel port 
  SendBit(setbit(VDD));      // set VDD to high
  if(OPTO)
  {
    SendBit(setbit(VDD2));     // set VDD to high
    SendBit(clearbit(TEST));   // pull test low
    SendBit(clearbit(DIRO));   // Ensure DIRO is high
  }

  textcolor(LIGHTGRAY); textbackground(BLACK);
  clrscr();
  printf("\nEEPROM Configuration, Reading and Writing \n");
  Calculate();
  BlockDia();
  Menu(Cpos);
  Cpos=CursorPos(Key);

  while((ch !='q')&&(ch !='Q'))
  {
    if (bioskey(1))
    {
      Key = bioskey(0);
      ch = Key&0xff;
      if (ch==0)
        Cpos=CursorPos(Key);

      if ((int)ch ==13)
      {
        GetPar(Cpos);
        Calculate();
        BlockDia();
        Menu(Cpos);
        MemCode();
      }
      if (ch=='l')
      {
        MemList();
        BlockDia();
        Menu(Cpos);
        ch=' ';
      }
      if (ch=='w')
      {
        Write2EE();
        BlockDia();
        Menu(Cpos);
        TrigReload();
        ch=' ';
      }
      if (ch=='r')
      {
        SQ_read();
        MemList();
        BlockDia();
        Menu(Cpos);
        ch=' ';
      }
      if (ch=='t')
      {
        TrigReload();
        ch=' ';
      }
    }
  }
}
